/* ************************************************************************* *
   LPEvol - (C) Copyright 1998 Robert Gasch (Robert_Gasch@peoplesoft.com)
	        http://www.peoplesoft.com/peoplepages/g/robert_gasch/index.htm

   Permission to use, copy, modify and distribute this software for any 
   purpose and without fee is hereby granted, provided that this copyright
   notice appear in all copies as well as supporting documentation. All
   work developed as a consequence of the use of this program should duly
   acknowledge such use.

   No representations are made about the suitability of this software for
   any purpose. This software is provided "as is" without express or implied 
   warranty.

   See the GNU General Public Licence for more information.
 * ************************************************************************* */


import java.lang.*;
import java.util.*;


/* *********************************************************************** */
/* **** LParserGA: Class to implement GA functions for LParser orgs ****** */
/* *********************************************************************** */
public class LParserGA extends GA
{
	public static final int		s_LIMIT_BRACKETS_OFF = 0;
	public static final int		s_LIMIT_BRACKETS_ALL = 1;
	public static final int		s_LIMIT_BRACKETS_LEFT = 2;
	public static final int		s_LIMIT_BRACKETS_RIGHT = 3;
	public static final int		s_AXIOM_SKIP = 1;		// offset for actual program instructions, 
	public static final int		s_CHROM_OFFSET = s_AXIOM_SKIP;		// offset for actual program instructions, 
	public static final int		s_EQUATION_PREFIX = 2;	// offset for actual program instructions, 
	public static final char 	s_PUSH_TOK='[';		// push token (LParser syntax)
	public static final char 	s_POP_TOK=']';		// pop token (LParser syntax)
	public static final int		d_max_brackets_diff=0;

	private LParserObj		d_LPO;		// LParserObj for which we do the crossover 
	private int			d_limitBrackets=s_LIMIT_BRACKETS_ALL;
	private float			d_append=(float)0;
	private float			d_appendReplaceProb=(float)0.5;
	private	int			d_chromOffset=s_CHROM_OFFSET;	// actual header length
	private	int			d_numberBreed=0;	// actual header length
	private float			d_numberBreedProb=(float)0.5;
	private boolean			d_crossBeforeNumber=true;

	/** Constructor: check valid LParserObj and assign local LPO */
	public LParserGA (LParserObj	LPO)
		{
		if (LPO == null)
			{
			System.err.println ("Error: _lso_==null in LParserGA ... Exiting");
			System.exit (1);
			}
		d_LPO=LPO;
		this.d_gene=this.d_chrom=0; 	// inherited from GA
		}

        /** setCrossBeforeNumber: set the crossBeforeNumber option */
	public void setCrossBeforeNumber (boolean crossBeforeNumber)
		{
		d_crossBeforeNumber = crossBeforeNumber;
		}

        /** setChromOffset: set the number of chroms we skip before breeding */
	public void setChromOffset (int coffset)
		{
		if (coffset < 0 || coffset > s_AXIOM_SKIP)
			return;
		d_chromOffset = coffset;
		}

        /** setNumberBreed: set the number of numeric parameters breed */
	public void setNumberBreed (int nbreed, float numberBreedProb)
		{
		if (nbreed < 0 || nbreed > s_AXIOM_SKIP)
			return;
		d_numberBreed = nbreed;

		if (numberBreedProb < 0 || numberBreedProb > 1) 
			return;
		d_numberBreedProb = numberBreedProb;
		}

        /** setLimitBrackets: set the limitBrackets option */
	public void setLimitBrackets (int limit_brackets)
		{
		if (limit_brackets >= s_LIMIT_BRACKETS_OFF  &&
		    limit_brackets <= s_LIMIT_BRACKETS_RIGHT)
			d_limitBrackets = limit_brackets;
		}

        /** setAppend: set the append percentages */
	public void setAppend (float append, float appendReplaceProb)
		{
		if (append >= 0 && append <= 1)
			d_append = append;
		if (appendReplaceProb > 0 && appendReplaceProb <1)
			d_appendReplaceProb = appendReplaceProb;
		}

	/** rndGlobalSetting: select random LParser global setting */
	public int rndGlobalSetting ()
		{
		int	t = (int)(JAC_GlobalRandom.nextFloat()*LParserObj.s_NUM_VALUES);

		if (t == LParserObj.s_NUM_VALUES) 	// random number was 1 
			t--;
		return (t);
		}

	/** isFit: returns if the org is fit or not */
	public boolean isFit ()
		{
		if (fitness() == 1)
			return (true);
		else
			return (false);
		}
	
	/** fitness: return the org's fitness on a scale of 0<=n<=1 */
	public float fitness ()
		{
		if (d_LPO.getSelected())
			return (1);
		else
			return (0);
		}

	/** rndChrom: select random chromosme from the program ones only */
	public int rndChrom ()
		{
		return (rndChrom(d_chromOffset));
		}

	/** rndChrom: select a random chromosme from all chromosomes. If 
	    offset < 3 we ignore it and pretend it's three (since the 
	    first three parameters are basically global parameters 
	    and as such not LParser systems which we can breed)
	**/
	public int rndChrom (int offset)
		{
		// only include the actual LParser program 
		if (offset >= 0 && offset <= s_AXIOM_SKIP)
			d_chrom=(int)(JAC_GlobalRandom.nextFloat()*(d_LPO.size()-offset)+offset);
		else
			return (-1);
		return (d_chrom);
		}
	
	/** rndGene: return Gene Crossover Point for a random chromosome */
	public int rndGene ()
		{
		return (rndGene(s_EQUATION_PREFIX, rndChrom()));
		}

	/** rndGene: return random crossover point for a random chromosome */
	public int rndGene (int offset)
		{
		return (rndGene(offset, rndChrom()));
		}

	/** rndGene: return random crossover Point for given chromosome */
	public int rndGene (int off, int chromosome)
		{
		final int	margin=1;
		int 		in_bracket_left=0, in_bracket_right=0;
		int		guess, corrected;
		String 		Cr = (String)d_LPO.elementAt(chromosome);

		// take a first guess, skip the "X=" characters 
		if (Cr.length() > margin*2+off)
			guess=(int)(JAC_GlobalRandom.nextFloat()*
					(Cr.length()-off-margin*2)+off+margin);
		else
			{
			guess=(int)(JAC_GlobalRandom.nextFloat()*Cr.length());
			if (guess == Cr.length() && guess > 0)
				guess--;
			}

		corrected = LParserLangUtil.fixNumberPoint (Cr, guess, d_crossBeforeNumber)-off;
		guess-=off;
		d_gene=corrected;

		if (JAC_GlobalDebug.debug() > 1)
			System.out.println ("rndGene: Guess=" + guess + 
				" corrected=" + corrected + 
				" data= " + Cr.substring(2) + ", "+ Cr.length());

		return (corrected);
		}

	/** breed: breed the elegible population */
	public LParserPop breed (LParserPop LPP)
		{
		return (breed (LPP, 0, true));
		}

	/** breed: breed the elegible population */
	public LParserPop breed (LParserPop LPP, int maxorgs)
		{
		return (breed (LPP, maxorgs, true));
		}

	/** breed: breed the elegible population, create maxorgs children. 
	    maxorgs==0 means that we create LPP.size() children 
	**/
	public LParserPop breed (LParserPop LPP, int maxorgs, boolean count)
		{
		Vector 		EPopIdx = new Vector (5, 5), 	// Elegible Population Index
				BreedPlan = null;		// Breeding Plan 
		int		t=0, t1=-1,
				lim = LPP.size();
		LParserPop	LPPC = new LParserPop ();	// LParser Child Population
		LParserObj	Parent1=null, Parent2=null; 	// LParser Parent objects

		// build an index of EPopIdx indexes
		for (int i=0; i<lim; i++)
			{
			LParserObj	LPO = (LParserObj)LPP.elementAt(i);

			if (LPO.getSelected ())
				EPopIdx.addElement (new Integer (i));
			}

		// if there are not enough elegible orgs, create 
		// a random breed list from the entire population 
		if (EPopIdx.size() < 2)
			for (int i=0; i<lim; i++)
				{
				t = (int)(JAC_GlobalRandom.nextFloat()*lim);

				// redo if 2nd index is same as 1st
				if (i%2 == 1 && t1 == t)
					t--;
				else
					{
					EPopIdx.addElement (new Integer (t));
					t1 = t;
					}
				}

		if (maxorgs > 0)
			lim = maxorgs;

		// create the breeding plan, returns a Vector 
		// of Integer subscripts, static function 
		BreedPlan = GA.breedLottery (EPopIdx, lim);

		for (int i=0; i<lim; i+=2)
			{
			if (JAC_GlobalDebug.verbose()) 
				System.out.println ("\nBreeding org " + i + 
							" + " + (i+1));

			// get the index, then get the LParserObj it points at
			Integer I1 = (Integer)BreedPlan.elementAt(i);
			Integer I2 = (Integer)BreedPlan.elementAt(i+1);
			Parent1 = (LParserObj)LPP.elementAt(I1.intValue());
			Parent2 = (LParserObj)LPP.elementAt(I2.intValue());

			// Sanity Check	
			// if (Parent1 == null || Parent2 == null)
			//	JAC_Bailout.bailout (new String (
			//		"Sanity Check failed: Parent1 = " + 
			//		Parent1 + " Parent2 = " + Parent2), -1);
			//	System.exit (1);
			//	}

			// advance generation count once per generation 
			if (i==0 && count)
				Parent1.nextGeneration();

			// breed Parent1 with Parent2, add results to 
			// LPPC. Passing i and lim enables us to correctly 
			// handle odd population sizes
			orgBreed (LPPC, i, lim, Parent1, Parent2);
			}

		//LPPC.setRenderMode (LPP.getRenderMode());

		// return LParserPop pointer
		return (LPPC);
		}

	/** Breed: breed LPO1 with LPO2. Passing the first 3 parameters 
	    to allow us to automate the addition to the population of 
	    the result of combining the parents.
	**/
	private int orgBreed (LParserPop LPPC, int c, int lim, 
			  LParserObj Parent1, LParserObj Parent2)
		{
		LParserObj	LPOChild1=null, LPOChild2=null, 	// LParser Child objects
				Children[]=null;

		// do the crossover 
		Children =  doCrossover (Parent1, Parent2);

		System.out.println ("YYYY " + (Parent1.childCount()));
		System.out.println ("YYYY " + (Parent2.childCount()));

		// get 1st child, record parent names and up counters
		LPOChild1 = (LParserObj)Children[0];
		LPOChild1.setName (Parent1.getName());
		LPOChild1.setParents (Parent1, Parent2);
		LPOChild1.nextChild ();

		// same for 2nd child
		LPOChild2 = (LParserObj)Children[1];
		LPOChild2.setName (Parent2.getName());
		LPOChild2.setParents (Parent2, Parent1);
		LPOChild2.nextChild ();

		// add to population 
		LPPC.add (LPOChild1);

		// if this is true, lim is odd and we're done	
		if (c == lim-1)
			;
		else
			// add to population 
			LPPC.add (LPOChild2);

		return (0);
		}

	/** doCrossover: do the crossover reproduction */
	private LParserObj[] doCrossover (LParserObj Parent1, LParserObj Parent2)
		{
		int		c1, c2,		// chromosome subscripts
				g1, g2, g=-1, 	// 2 child and 1 global gene subscripts
				sbal1, sbal2;	// child string balances
		String 		Cs1, Cs2,	// Chromosome String values
				Final1, Final2,	// the line after the prefix (after '=')
				New1, New2,	// new organisms after crossover 
				Pre1=null, 	// the "X=" prefix of the line (up to '=')
				Pre2=null;	// the "X=" prefix of the line (up to '=')
		LParserObj	Child1, Child2;	// Child objects
		LParserObj	CLPO[] = new LParserObj[2]; // Final Array return type 
		LParserGA	Pga1, Pga2; 	// Parent LParserGA objects
		LParserLangUtil	Lplu=null;

		if (JAC_GlobalDebug.verbose())
			System.out.println ("Initiating crossover between " + 
				Parent1.fullName() + " and " + Parent2.fullName());

		// select random genes 1st parent
		Pga1 = new LParserGA (Parent1);
		c1 = Pga1.rndChrom ();
		Cs1 = (String)Parent1.elementAt (c1);
		if (c1 == 0)
			{
			Final1=Cs1;
			g1=Pga1.rndGene(0, c1);
			}
		else
			{
			Pre1=Cs1.substring (0, s_EQUATION_PREFIX);
			Final1=Cs1.substring (s_EQUATION_PREFIX);
			g1=Pga1.rndGene(s_EQUATION_PREFIX, c1);
			}

		// select random genes 2nd parent
		Pga2 = new LParserGA (Parent2);
		c2 = Pga2.rndChrom ();
		Cs2 = (String)Parent2.elementAt (c2);
		if (c2 == 0)
			{
			Final2=Cs2;
			g2=Pga2.rndGene(0, c2);
			}
		else
			{
			Pre2=Cs2.substring (0, s_EQUATION_PREFIX);
			Final2=Cs2.substring (s_EQUATION_PREFIX);
			g2=Pga2.rndGene(s_EQUATION_PREFIX, c2);
			}

		// g1 & g2 are index subscripts. Since we want to include 
		// them in the string we up the count one to account for 
		// the way String.substring works (unless gx == last character)
		if (g1 < Final1.length()-1)
			g1++;
		if (g2 < Final2.length()-1)
			g2++;

		// Construct the final strings. Balance the substrings 
		// to ensure the we do't accumulate brackets. Get the 
		// appropriate substring and reduce it's imbalance 
		// to the appropriate level. Then reconstruct the 
		// strings which are then passed to Crossover
		if (d_limitBrackets != s_LIMIT_BRACKETS_OFF)
			{
			int		bal;
			String		Empty = new String ();
			String 		S1_1=null, S1_2=null, 
					S2_1=null, S2_2=null,
					F1_1=Empty, F1_2=Empty,
					F2_1=Empty, F2_2=Empty;

			if (d_limitBrackets == s_LIMIT_BRACKETS_ALL ||
			    d_limitBrackets == s_LIMIT_BRACKETS_LEFT)
				{
				S1_1 = Final1.substring(0,g1);
				Lplu = new LParserLangUtil (S1_1);
				F1_1 = Lplu.ppBalance (0, false);

				S2_1 = Final2.substring(0,g2);
				Lplu = new LParserLangUtil (S2_1);
				F2_1 = Lplu.ppBalance (0, false);

				g1 = g1 - (g1-F1_1.length());
				g2 = g2 - (g2-F2_1.length());
				}
			else
				{
				F1_1 = Final1.substring(0,g1);
				F2_1 = Final2.substring(0,g2);
				}

			if (d_limitBrackets == s_LIMIT_BRACKETS_ALL ||
			    d_limitBrackets == s_LIMIT_BRACKETS_RIGHT)
				{
				S1_2 = Final1.substring(g1,Final1.length());
				Lplu = new LParserLangUtil (S1_2);
				F1_2 = Lplu.ppBalance (0, false);

				S2_2 = Final2.substring(g2,Final2.length());
				Lplu = new LParserLangUtil (S2_2);
				F2_2 = Lplu.ppBalance (0, false);

				g1 = g1 - (g1-F1_1.length());
				g2 = g2 - (g2-F2_1.length());
				}
			else
				{
				F1_2 = Final1.substring(g1,Final1.length());
				F2_2 = Final2.substring(g2,Final2.length());
				}

			Final1 = new String (F1_1 + F1_2);
			Final2 = new String (F2_1 + F2_2);
			}

		//System.out.println ("CrossOver g=" + g + " g1=" + g1 + " c1=" + c1 + " data=" + Cs1);
		//System.out.println ("CrossOver g=" + g + " g2=" + g2 + " c2=" + c2 + " data=" + Cs2);

		// do the actual crossover !!!!!!!!!!!!!
		String [] children;
		children = GA.crossover (Final1, Final2, g1, g2);

		// construct full child chromosome string values and fix '[' and ']'
		// see if c1 or c2 reference the axiom, leave out
		// the left-hand side of the LParser equation if they do. 
		Lplu = new LParserLangUtil (children[0]);
		if (c1 == 0)
			New1=Lplu.ppBalance ();
		else
			New1=Pre1.concat (Lplu.ppBalance());

		Lplu = new LParserLangUtil (children[1]);
		if (c2 == 0)
			New2=Lplu.ppBalance ();
		else
			New2=Pre2.concat (Lplu.ppBalance ());

		// bail out if we broke the LParser stack symmetry. This 
		// should never happen and means something is wrong with 
		// the code. Should eventually be commented out. 
		int bal;
		if ((bal=JAC_StringUtil.stringBalance(New1, '[', ']')) != 0)
			JAC_Bailout.bailout (new String ("Sanity Check Failed in LParserGA::DoCrossover(): New1 [] balance=" 
					+ bal + "\nLine=" + New1), -1);
		if ((bal=JAC_StringUtil.stringBalance(New2, '[', ']')) != 0)
			JAC_Bailout.bailout (new String ("Sanity Check Failed in LParserGA::DoCrossover(): New2 [] balance=" 
					+ bal + "\nLine=" + New2), -1);

		// bail out: broken LParser number format. Same as above.
		if ((bal=JAC_StringUtil.stringBalance(New1, '(', ')')) != 0)
			JAC_Bailout.bailout (new String ("Sanity Check Failed in LParserGA::DoCrossover(): New1 () balance=" 
					+ bal + "\nLine=" + New1), -1);
		if ((bal=JAC_StringUtil.stringBalance(New2, '(', ')')) != 0)
			JAC_Bailout.bailout (new String ("Sanity Check Failed in LParserGA::DoCrossover(): New2 () balance=" 
					+ bal + "\nLine=" + New2), -1);

		// build return objects
		Child1=(LParserObj)Parent1.clone();
		Child2=(LParserObj)Parent2.clone();

		if (d_append > 0)
			{
			// append new chromosome to org
			if (d_append >= JAC_GlobalRandom.nextFloat())
				appendChrom (Child1, New1, 
					d_chromOffset, d_appendReplaceProb);
			if (d_append >= JAC_GlobalRandom.nextFloat())
				appendChrom (Child2, New2, 
					d_chromOffset, d_appendReplaceProb);
			}
		else
			{
			// overwrite original chromosome with child chromosome
			Child1.setElementAt (New1, c1);
			Child2.setElementAt (New2, c2);
			}

		if (d_numberBreed > 0)
			crossoverNumeric (Child1, Child2, 
					d_numberBreed, d_numberBreedProb);

		CLPO[0]=Child1;
		CLPO[1]=Child2;

		if (JAC_GlobalDebug.verbose())
			System.out.println ("");

		return (CLPO);
		}

	/** crossoverNumeric: crossover the numeric lparser parameters. 
	    If anybody has a better idea on how to do the numeric 
	    crossover, please let me know. Maybe a bit-level crossover 
	    would be an idea? 
	 **/
	
	private static void crossoverNumeric (LParserObj LPO1, LParserObj LPO2, 
					      int nbreed, float numberBreedProb)
		{
		// handle global lparser values 
		float		o1, o2, n1, n2;

		for (int i=0; i<nbreed; i++)
			{
			if (numberBreedProb > JAC_GlobalRandom.nextFloat())
				{
				o1 = LPO1.getValue (i);
				o2 = LPO2.getValue (i);
				if (LPEOptions.s_numberBreedInt)
					{
					n1=(float)((int)((o1+o2)/2*(JAC_GlobalRandom.nextFloat()+0.5)));
					n2=(float)((int)((o1+o2)/2*(JAC_GlobalRandom.nextFloat()+0.5)));
					}
				else
					{
					n1=(float)((o1+o2)/2*(JAC_GlobalRandom.nextFloat()+0.5));
					n2=(float)((o1+o2)/2*(JAC_GlobalRandom.nextFloat()+0.5));
					}
				LPO1.setValue (i, n1);
				LPO1.setValue (i, n2);
				}
			}
		}

	/** appendChrom: append Chrom to LSO and give it references */
	private static void appendChrom (LParserObj LSO, String Chrom, 
					 int chromOffset, float appendReplaceProb)
		{
		int	olim = LSO.size(),
			lim = olim-chromOffset,
			winner;
		char	mc =' ', 
			winnerName, 
			insertName;
		char[]	let1 = new char[lim];	// 1st letter in a line
		String	chrom;

		// build a 1st letter index and find max X= name 
		for (int i=chromOffset; i<olim; i++)
			{
			char 	tc;

			tc=((String)LSO.elementAt(i)).charAt(0);
			let1[i-chromOffset]=tc;
			if (tc > mc)
				mc = tc;
			}

		// guess the letter to partially replace 
		winner = (int)(JAC_GlobalRandom.nextFloat()*(olim-chromOffset)+chromOffset);
		if (winner == lim)
			winner--;
		winnerName = let1[winner];
		int t = (int)mc;
		insertName = (char)++t;

		// build chromosome as we want to insert it, advance last 
		// used name letter
		chrom = new String (insertName + Chrom.substring(1));

		for (int i=0; i<lim; i++)
			{
			String		Line = (String)LSO.elementAt(i+chromOffset);

			if (Line.charAt(0) != winnerName)
			    LSO.setElementAt(
				JAC_StringUtil.pStringReplace (Line, winnerName, 
					insertName, appendReplaceProb), 
					i+chromOffset);
			}

		if (JAC_GlobalDebug.debug() > 0)
			System.out.println ("Appending " + chrom + " to " + LSO.getName());

		LSO.addElement (chrom);
		}
}

