/* ************************************************************************* *
   PovUtil - (C) Copyright 1998 Robert Gasch (Robert_Gasch@peoplesoft.com)
	         http://www.peoplesoft.com/peoplepages/g/robert_gasch/index.htm

   Permission to use, copy, modify and distribute this software for any 
   purpose and without fee is hereby granted, provided that this copyright
   notice appear in all copies as well as supporting documentation. All
   work developed as a consequence of the use of this program should duly
   acknowledge such use.

   No representations are made about the suitability of this software for
   any purpose. This software is provided "as is" without express or implied 
   warranty.

   See the GNU General Public Licence for more information.
 * ************************************************************************* */


import java.lang.*;
import java.io.*;
import java.util.*;


/* ************************************************************************** */
/* ***** PovMaxVector: abstract class to serve as a base for a specific ***** */
/* ***** kind of PovMaxVector. Provides storage and comarison routines  ***** */
/* ************************************************************************** */
public abstract class PovMaxVector extends Vector implements Runnable 
{
	protected JAC_FileReader	d_Bfr=null;
	protected final int		d_dotinterval=250;
	protected int			d_lnumber=0;		// line number
	protected int			d_maxkeys=0;		// unlimited
	protected String		d_Fname;		// filename we're parsing
	protected PovVector		d_PVMax,
					d_PVMin;

	private Thread			d_Thread;

	/** constructor: open and process file without limits */
	public PovMaxVector (String Fname) throws FileNotFoundException, 
						  IOException
		{
		this (Fname, 0);
		}

	/** constructor: open file, establish keys and file max key vectors */
	public PovMaxVector (String Fname, int maxk) throws 
						     FileNotFoundException, 
						     IOException
		{
		super (5, 5);
		d_Fname = new String (Fname);
		d_maxkeys = maxk;
		d_Bfr=new JAC_FileReader (d_Fname);
		d_PVMin=new PovVector ("Min");
		d_PVMin.setValues (Double.POSITIVE_INFINITY, Double.POSITIVE_INFINITY, Double.POSITIVE_INFINITY);
		d_PVMax=new PovVector ("Max");
		d_PVMax.setValues (Double.NEGATIVE_INFINITY, Double.NEGATIVE_INFINITY, Double.NEGATIVE_INFINITY);
		}

	/** run: thread startup code */
	public void run ()
		{
		parse ();
		if (JAC_GlobalDebug.verbose() == true)
			System.out.println ("Finished Parsing " + d_Fname + ": " + d_lnumber + " lines");

		try { d_Bfr.close(); }
		catch (IOException e) {
			System.out.println ("Caught IOException in PovMaxVector::run() while closing file ...");
			}
		}

	/** merge: merge this PovMaxVector with previously existing ones */
	public void merge (PovMaxVector PMV)
		{
		double[]	d;

		d = PMV.getMinValues ();
		d_PVMin.keepSmaller (d[0], d[1], d[2]);

		d = PMV.getMaxValues ();
		d_PVMax.keepBigger (d[0], d[1], d[2]);
		}

	/** Print: print all the PovMaxVectors stored */
	public void print ()
		{
		double[]	d;

		System.out.println ("PovMaxVector:");
		d = d_PVMin.getValues ();
		System.out.println ("Min xyz: " + d[0] + ", " + d[1] + ", " + d[2]);
		d = d_PVMax.getValues ();
		System.out.println ("Max xyz: " + d[0] + ", " + d[1] + ", " + d[2]);
		}

	/** getMaxValues: return the max vectors */
	public double[] getMaxValues ()
		{
		return (d_PVMax.getValues());
		}

	/** getMinValues: return the min vectors */
	public double[] getMinValues ()
		{
		return (d_PVMin.getValues());
		}

	/** addKey: establish a new Key */
	public PovVector addKey (String Key)
		{
		PovVector	Pv=null;

		if (d_maxkeys > 0)
			if (this.size() < d_maxkeys)
				Pv = new PovVector (Key);
			else
				{
				System.err.println ("Warning: PovMaxVector::addKey() cancelled ... Vector is full\n");
				return (null);
				}
		else
				Pv = new PovVector (Key);

		this.addElement (Pv);
		return (Pv);
		}

	/** parse: read d_Bfr and process it */
	public abstract void parse ();

	public void setThread (Thread T)
		{
		if (T == null)
			{
			System.err.println ("Sanity Check Failed in PovMaxVector::setThread() ... T==null\n");
			System.exit (1);
			}

		d_Thread = T;
		}

	public void yield ()
		{
		if (d_Thread != null)
			d_Thread.yield ();
		}

	/** parse_vector: parse a string and return the XYZ array of Float[] */
	protected Vector parse_vector (String Line)
		{
		int		lim=1;
		Double 		D1, D2, D3;
		String		S1, S2, S3;
		Vector		VDouble = new Vector (3, 3);
		StringTokenizer	ST = new StringTokenizer (Line, "<,>");

		if (!ST.hasMoreTokens())
			return (null);

		if (Line.startsWith("triangle"))
			lim=3;

		ST.nextToken();
		for (int i=0; i<lim; i++)
			{
			try {
			    S1=ST.nextToken();
			    S2=ST.nextToken();
			    S3=ST.nextToken();
			    D1=Double.valueOf (S1);
			    D2=Double.valueOf (S2);
			    D3=Double.valueOf (S3);
			    VDouble.addElement (D1);
			    VDouble.addElement (D2);
			    VDouble.addElement (D3); 
			    }
			catch (NumberFormatException e) {
			    System.err.println ("\nCaught exception in PovMaxVector::parse_vector():" + e + " while parsing " + d_Fname + " (Line  " + d_lnumber + "):\n" + Line);
			    return (null);
			    }
			catch (NoSuchElementException e) {
			    System.err.println ("\nCaught exception in PovMaxVector::parse_vector():" + e + " while parsing " + d_Fname + " (Line  " + d_lnumber + "):\n" + Line);
			    return (null);
			    }
			}

		return (VDouble);
		}
}

